from __future__ import absolute_import, division, print_function, unicode_literals


class CliCommand(object):
    @classmethod
    def register_parser(cls, subparser_root):
        """
        Registers the command's parser arguments.

        :return: the subparser
        """
        raise NotImplementedError()

    @classmethod
    def run_command(cls, args):
        raise NotImplementedError()



from __future__ import absolute_import, division, print_function, unicode_literals

from ._formats import ImageFormat
from ._collection import Collection
from ._tile import Tile
from ._tileset import TileSet
from .io import Reader, Writer, v0_0_0


__all__ = [
    Collection,
    ImageFormat,
    Reader,
    Tile,
    TileSet,
    Writer,
    v0_0_0,
]



import enum


def format_tile_coordinates(tile_dimensions):
    """
    Given a dictionary mapping keys to values, where the keys may either be strings or enums, and
    the values may either be a scalar number or an iterable of two scalar numbers, return a new
    dictionary with the same contents, except the keys are converted to strings, and the value is
    transformed as per the following rules:

    scalar_number -> (scalar_number, scalar_number)
    [scalar_number_0, scalar_number_1] -> (scalar_number_0, scalar_number_1)
    """
    result = dict()
    for name, value in tile_dimensions.items():
        key = _str_or_enum_to_str(name)
        try:
            iter(value)
            value = tuple(value)
            if len(value) == 2:
                result[key] = value
            else:
                raise ValueError("Not a valid input")
        except TypeError:
            result[key] = (value, value)
    return result


def format_tile_indices(tile_dimensions):
    """
    Given a dictionary mapping keys to values, where the keys may either be strings or enums,
    return a new dictionary with the same contents, except the keys are converted to strings.
    """
    result = dict()
    for name, value in tile_dimensions.items():
        result[_str_or_enum_to_str(name)] = value
    return result


def format_tileset_dimensions(tileset_dimensions):
    """
    Given an iterable of strings or enums, return a frozenset consisting of the same values, except
    all converted to strings.
    """
    return frozenset(
        _str_or_enum_to_str(tileset_dimension)
        for tileset_dimension in tileset_dimensions)


def format_tileset_shape(d):
    """
    Given a dictionary mapping keys to values, where the keys may either be strings or enums, return
    a new dictionary where the keys are all converted to strings.
    """
    result = dict()
    for name, value in d.items():
        result[_str_or_enum_to_str(name)] = value
    return result


def _str_or_enum_to_str(value):
    """
    Given a scalar value or an enum, return the scalar value if it's a scalar value, or the enum's
    value field if it's an enum.
    """
    if isinstance(value, enum.Enum):
        return value.value
    else:
        return value



import posixpath

from six.moves import urllib


def pathsplit(url):
    parsed = urllib.parse.urlparse(url)
    return (
        urllib.parse.urlunparse(
            (parsed.scheme,
             parsed.netloc,
             posixpath.dirname(parsed.path),
             parsed.params,
             parsed.query,
             parsed.fragment),

        ),
        posixpath.basename(parsed.path),
    )


def pathjoin(url, *segments):
    parsed = urllib.parse.urlparse(url)
    return urllib.parse.urlunparse(
        (parsed.scheme,
         parsed.netloc,
         posixpath.join(parsed.path, *segments),
         parsed.params,
         parsed.query,
         parsed.fragment))



import contextlib
import socket

from tests.utils.contextchild import ContextualChildProcess

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from .tempdir import TemporaryDirectory


__all__ = [
    TemporaryDirectory,
    ContextualChildProcess,
    'unused_tcp_port',
    'build_skeleton_manifest'
]


def unused_tcp_port():
    """
    Return an unused TCP port.
    """
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('127.0.0.1', 0))
        return sock.getsockname()[1]


def build_skeleton_manifest():
    """
    Returns a 0.0.0 formatted manifest with no tiles.
    """
    return {
        "version": "0.0.0",
        "dimensions": [
            "x",
            "y",
            "hyb",
            "ch"
        ],
        "shape": {
            "hyb": 1,
            "ch": 1
        },
        "tiles": []
    }



import codecs
import hashlib
import json
import os
import tempfile
import unittest

import numpy as np
import skimage.io

import slicedimage
from slicedimage import ImageFormat
from tests.utils import TemporaryDirectory, build_skeleton_manifest

baseurl = "file://{}".format(os.path.abspath(os.path.dirname(__file__)))


class TestWrite(unittest.TestCase):
    def test_write_tileset(self):
        image = slicedimage.TileSet(
            ["x", "y", "ch", "hyb"],
            {'ch': 2, 'hyb': 2},
            (100, 100),
        )

        for hyb in range(2):
            for ch in range(2):
                tile = slicedimage.Tile(
                    {
                        'x': (0.0, 0.01),
                        'y': (0.0, 0.01),
                    },
                    {
                        'hyb': hyb,
                        'ch': ch,
                    },
                )
                tile.numpy_array = np.zeros((100, 100))
                tile.numpy_array[hyb, ch] = 1
                image.add_tile(tile)

        with TemporaryDirectory() as tempdir, \
                tempfile.NamedTemporaryFile(suffix=".json", dir=tempdir) as partition_file:
            partition_doc = slicedimage.v0_0_0.Writer().generate_partition_document(
                image, partition_file.name)
            writer = codecs.getwriter("utf-8")
            json.dump(partition_doc, writer(partition_file))
            partition_file.flush()

            basename = os.path.basename(partition_file.name)
            baseurl = "file://{}".format(os.path.dirname(partition_file.name))

            loaded = slicedimage.Reader.parse_doc(basename, baseurl)

            for hyb in range(2):
                for ch in range(2):
                    tiles = [_tile
                             for _tile in loaded.tiles(
                                 lambda tile: (tile.indices['hyb'] == hyb and
                                               tile.indices['ch'] == ch))]

                    self.assertEqual(len(tiles), 1)

                    expected = np.zeros((100, 100))
                    expected[hyb, ch] = 1

                    self.assertEqual(tiles[0].numpy_array.all(), expected.all())
                    self.assertIsNotNone(tiles[0].sha256)

    def test_write_collection(self):
        image = slicedimage.TileSet(
            ["x", "y", "ch", "hyb"],
            {'ch': 2, 'hyb': 2},
            (100, 100),
        )

        for hyb in range(2):
            for ch in range(2):
                tile = slicedimage.Tile(
                    {
                        'x': (0.0, 0.01),
                        'y': (0.0, 0.01),
                    },
                    {
                        'hyb': hyb,
                        'ch': ch,
                    },
                )
                tile.numpy_array = np.zeros((100, 100))
                tile.numpy_array[hyb, ch] = 1
                image.add_tile(tile)
        collection = slicedimage.Collection()
        collection.add_partition("fov002", image)

        with TemporaryDirectory() as tempdir, \
                tempfile.NamedTemporaryFile(suffix=".json", dir=tempdir) as partition_file:
            partition_doc = slicedimage.v0_0_0.Writer().generate_partition_document(
                collection, partition_file.name)
            writer = codecs.getwriter("utf-8")
            json.dump(partition_doc, writer(partition_file))
            partition_file.flush()

            basename = os.path.basename(partition_file.name)
            baseurl = "file://{}".format(os.path.dirname(partition_file.name))

            loaded = slicedimage.Reader.parse_doc(basename, baseurl)

            for hyb in range(2):
                for ch in range(2):
                    tiles = [_tile
                             for _tile in loaded.tiles(
                                 lambda tile: (tile.indices['hyb'] == hyb and
                                               tile.indices['ch'] == ch))]

                    self.assertEqual(len(tiles), 1)

                    expected = np.zeros((100, 100))
                    expected[hyb, ch] = 1

                    self.assertEqual(tiles[0].numpy_array.all(), expected.all())
                    self.assertIsNotNone(tiles[0].sha256)

    def test_checksum_on_write(self):
        """
        Generate a tileset consisting of a single TIFF tile.  Load it and then write it back out
        as a numpy tile, which should be written with different checksums.  Then verify that the
        numpy version can load without an error.
        """
        # write the tiff file
        with TemporaryDirectory() as tempdir:
            data = np.random.randint(0, 65535, size=(100, 100), dtype=np.uint16)
            file_path = os.path.join(tempdir, "tile.tiff")
            skimage.io.imsave(file_path, data, plugin="tifffile")
            with open(file_path, "rb") as fh:
                checksum = hashlib.sha256(fh.read()).hexdigest()

            manifest = build_skeleton_manifest()
            manifest['tiles'].append(
                {
                    "coordinates": {
                        "x": [
                            0.0,
                            0.0001,
                        ],
                        "y": [
                            0.0,
                            0.0001,
                        ]
                    },
                    "indices": {
                        "hyb": 0,
                        "ch": 0,
                    },
                    "file": "tile.tiff",
                    "format": "tiff",
                    "sha256": checksum,
                },
            )
            with open(os.path.join(tempdir, "tileset.json"), "w") as fh:
                fh.write(json.dumps(manifest))

            image = slicedimage.Reader.parse_doc(
                "tileset.json",
                "file://{}".format(tempdir),
                allow_caching=False,
            )

            with TemporaryDirectory() as output_tempdir, \
                    tempfile.NamedTemporaryFile(
                        suffix=".json", dir=output_tempdir) as partition_file:
                partition_doc = slicedimage.v0_0_0.Writer().generate_partition_document(
                    image, partition_file.name)

                writer = codecs.getwriter("utf-8")
                json.dump(partition_doc, writer(partition_file))
                partition_file.flush()

                basename = os.path.basename(partition_file.name)
                baseurl = "file://{}".format(os.path.dirname(partition_file.name))

                loaded = slicedimage.Reader.parse_doc(basename, baseurl)

                loaded.tiles()[0].numpy_array

    def test_write_tiff(self):
        image = slicedimage.TileSet(
            dimensions=["x", "y", "ch", "hyb"],
            shape={'ch': 2, 'hyb': 2},
            default_tile_shape=(100, 100),
        )

        for hyb in range(2):
            for ch in range(2):
                tile = slicedimage.Tile(
                    coordinates={
                        'x': (0.0, 0.01),
                        'y': (0.0, 0.01),
                    },
                    indices={
                        'hyb': hyb,
                        'ch': ch,
                    },
                )
                tile.numpy_array = np.zeros((100, 100), dtype=np.uint32)
                tile.numpy_array[hyb, ch] = 1
                image.add_tile(tile)

        with TemporaryDirectory() as tempdir, \
                tempfile.NamedTemporaryFile(suffix=".json", dir=tempdir) as partition_file:
            # create the tileset and save it.
            partition_doc = slicedimage.v0_0_0.Writer().generate_partition_document(
                image, partition_file.name, tile_format=ImageFormat.TIFF)
            writer = codecs.getwriter("utf-8")
            json.dump(partition_doc, writer(partition_file))
            partition_file.flush()

            # construct a URL to the tileset we wrote, and load the tileset.
            basename = os.path.basename(partition_file.name)
            baseurl = "file://{}".format(os.path.dirname(partition_file.name))
            loaded = slicedimage.Reader.parse_doc(basename, baseurl)

            # compare the tiles we loaded to the tiles we set up.
            for hyb in range(2):
                for ch in range(2):
                    tiles = [_tile
                             for _tile in loaded.tiles(
                                 lambda tile: (tile.indices['hyb'] == hyb and
                                               tile.indices['ch'] == ch))]

                    self.assertEqual(len(tiles), 1)

                    expected = np.zeros((100, 100), dtype=np.uint32)
                    expected[hyb, ch] = 1

                    self.assertEqual(tiles[0].numpy_array.all(), expected.all())
                    self.assertIsNotNone(tiles[0].sha256)



from __future__ import absolute_import, division, print_function, unicode_literals

from slicedimage import Reader, Writer
from slicedimage.io import resolve_path_or_url
from ._base import CliCommand


class ChecksumCommand(CliCommand):
    @classmethod
    def register_parser(cls, subparser_root):
        checksum_command = subparser_root.add_parser(
            "checksum",
            help="Read a partition file and add missing checksums.")
        checksum_command.add_argument(
            "in_url",
            help="URL for the source partition file")
        checksum_command.add_argument(
            "out_path",
            help="Path to write partition file with checksums")
        checksum_command.add_argument(
            "--pretty",
            action="store_true",
            help="Pretty-print the output file")

        return checksum_command

    @classmethod
    def run_command(cls, args):
        _, name, baseurl = resolve_path_or_url(args.in_url)
        slicedimage = Reader.parse_doc(name, baseurl)

        Writer.write_to_path(
            slicedimage,
            args.out_path,
            pretty=args.pretty,
            tile_opener=fake_file_opener,
            tile_writer=identity_writer)


def fake_file_opener(partition_path, tile, ext):
    class null_file_handle(object):
        def __init__(self, name):
            self.name = name

        def write(self, data):
            return len(data)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    return null_file_handle(tile._file_or_url)


def identity_writer(tile, fh):
    assert tile._source_fh_contextmanager is not None
    with tile._source_fh_contextmanager as sfh:
        fh.write(sfh.read())
    return tile.tile_format



from __future__ import absolute_import, division, print_function, unicode_literals


from ._typeformatting import format_tileset_dimensions, format_tileset_shape


class TileSet(object):
    def __init__(
            self,
            dimensions,
            shape,
            default_tile_shape=None,
            default_tile_format=None,
            extras=None):
        self.dimensions = format_tileset_dimensions(dimensions)
        self.shape = format_tileset_shape(shape)
        self.default_tile_shape = None if default_tile_shape is None else tuple(default_tile_shape)
        self.default_tile_format = default_tile_format
        self.extras = {} if extras is None else extras
        self._tiles = []

        self._discrete_dimensions = set()

    def validate(self):
        raise NotImplementedError()

    def add_tile(self, tile):
        self._tiles.append(tile)

    def tiles(self, filter_fn=lambda _: True):
        """
        Return the tiles in this tileset.  If a filter_fn is provided, only the tiles for which
        filter_fn returns True are returned.
        """
        return list(filter(filter_fn, self._tiles))

    def get_dimension_shape(self, dimension_name):
        return self.shape[dimension_name]



from __future__ import absolute_import, division, print_function, unicode_literals

from ._base import ChecksumValidationError
from ._caching import CachingBackend
from ._disk import DiskBackend
from ._http import HttpBackend


__all__ = [
    CachingBackend,
    ChecksumValidationError,
    DiskBackend,
    HttpBackend,
]
